/* This file is part of the Palabos library.
 * Copyright (C) 2009, 2010 Jonas Latt
 * E-mail contact: jonas@lbmethod.org
 * The most recent release of Palabos can be downloaded at
 * <http://www.lbmethod.org/palabos/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/** \file
 * Helper functions for domain initialization -- header file.
 */
#ifndef INITIALIZATION_FUNCTIONAL_3D_HH
#define INITIALIZATION_FUNCTIONAL_3D_HH

#include "initializationFunctional3D.h"
#include "core/plbDebug.h"
#include "core/util.h"
#include "core/blockStatistics.h"
#include "latticeBoltzmann/momentTemplates.h"
#include "latticeBoltzmann/geometricOperationTemplates.h"
#include "finiteDifference/fdStencils1D.h"
#include "atomicBlock/atomicBlock3D.h"
#include "atomicBlock/blockLattice3D.h"
#include "atomicBlock/dataField3D.h"
#include "thrombosisLattices3D.h"
#include "nodeNatureTemplates.h"
#include <cmath>
#include <limits>

namespace plb {

/* ******** IniWallFlagsFunctional3D ****************************************** */


template<typename T1, template<class U> class Descriptor1>
IniFreeSurfaceFlagFunctional3D<T1,Descriptor1>::
    IniFreeSurfaceFlagFunctional3D(plint iniWaterLevel_, Box3D inletPouringLiquid_) :
        iniWaterLevel(iniWaterLevel_), inletPouringLiquid(inletPouringLiquid_)
{  }

template<typename T1, template<class U> class Descriptor1>
void IniFreeSurfaceFlagFunctional3D<T1,Descriptor1>::processGenericBlocks (
        Box3D domain,std::vector<AtomicBlock3D*> atomicBlocks )
{
    enum { nodeNature = Descriptor1<T1>::ExternalField::nodeNatureBeginsAt,
           time = Descriptor1<T1>::ExternalField::timeBeginsAt
    };

    using namespace freeSurfaceFlag;



     BlockLattice3D<T1,Descriptor1>& lattice =
            *dynamic_cast<BlockLattice3D<T1,Descriptor1> *>(atomicBlocks[0]);

    ScalarField3D<int>& flag =
            *dynamic_cast<ScalarField3D<int> *>(atomicBlocks[1]);

    ScalarField3D<int>& bounceBackMatrix =
            *dynamic_cast<ScalarField3D<int> *>(atomicBlocks[2]);


    Dot3D offset = computeRelativeDisplacement(lattice, bounceBackMatrix);
    Dot3D offsetFlag = computeRelativeDisplacement(lattice, flag);
    Dot3D absoluteOffset = lattice.getLocation();

    plint counterWet = 0;
    plint counterNotBounceBack = 0;
    for (plint iX=domain.x0; iX<=domain.x1; ++iX) {
        plint iX_bb = iX + offset.x;
	    plint flagX = iX + offsetFlag.x;
        for (plint iY=domain.y0; iY<=domain.y1; ++iY) {
            plint iY_bb = iY + offset.y;
	        plint flagY = iY + offsetFlag.y;
            for (plint iZ=domain.z0; iZ<=domain.z1; ++iZ) {
                plint iZ_bb = iZ + offset.z;
                plint flagZ = iZ + offsetFlag.z;
		        plint iZ_abs = iZ + absoluteOffset.z;

                if ( bounceBackMatrix.get(iX_bb,iY_bb,iZ_bb) == 0 ) //if NOT bounceback
                {
                    counterNotBounceBack ++;
                    if (iZ_abs <= iniWaterLevel )
                    {
                        flag.get(flagX,flagY,flagZ) = (int)freeSurfaceFlag::fluid;
                        counterWet++;
                    }
                    else
                    {
                        flag.get(flagX,flagY,flagZ) = (int)freeSurfaceFlag::empty;
                    }
                    if (iX == inletPouringLiquid.x0 && iZ_abs <= inletPouringLiquid.z1 )
                    {
                        flag.get(flagX,flagY,flagZ) = (int)freeSurfaceFlag::fluid;
                    }
                }
                else // if bounceback
                {
                    flag.get(flagX,flagY,flagZ) = (int)freeSurfaceFlag::wall;
                    if ( bounceBackMatrix.get(iX_bb,iY_bb,iZ_bb) == 1 )
                    {
                        *(lattice.get(iX,iY,iZ).getExternal(nodeNature)) = NodeNature::erodableWall;
                    }
                    else if( bounceBackMatrix.get(iX_bb,iY_bb,iZ_bb) == 2 )
                    {
                        *(lattice.get(iX,iY,iZ).getExternal(nodeNature)) = NodeNature::notErodableWall;
                    }
                    else
                    {
                        PLB_ASSERT(false);
                        //pcout << "We have a problem with bounceBackFlagMatrix: its value is not equal to 0 or 1 or 2" << std::endl;
                    }
		        }
            }
        }
    }
    // pcout << "Number of wet cells (excluding inlet) = " << counterWet << std::endl;
    // pcout << "Number of not bounce-back cells = " << counterNotBounceBack << std::endl;
    // pcout << "iZ min, iZ max, iniWaterLevel = " <<  domain.z0 + absoluteOffset.z << " " << domain.z1 + absoluteOffset.z << " " << iniWaterLevel << std::endl;
}

template<typename T1, template<class U> class Descriptor1>
IniFreeSurfaceFlagFunctional3D<T1,Descriptor1>* IniFreeSurfaceFlagFunctional3D<T1,Descriptor1>::clone() const {
    return new IniFreeSurfaceFlagFunctional3D<T1,Descriptor1>(*this);
}

template<typename T1, template<class U> class Descriptor1>
void IniFreeSurfaceFlagFunctional3D<T1,Descriptor1>::getTypeOfModification (
        std::vector<modif::ModifT>& modified ) const
{
    modified[0] = modif::dataStructure;
    // Full dynamics object must be recreated, because this data processor
    //   re-attributes a new dynamics and acts on the bulk only.
    modified[1] = modif::dataStructure;
    modified[2] = modif::nothing;
}


/////////////////////////////////////////////  ***********************


template<typename T, template<class U> class Descriptor>
DefineBulkStatusFunctional3D<T,Descriptor>::
        DefineBulkStatusFunctional3D(plint erodableWallLimit_):
        erodableWallLimit(erodableWallLimit_)
{ }


template<typename T, template<class U> class Descriptor>
void DefineBulkStatusFunctional3D<T,Descriptor>::process(
    Box3D domain, BlockLattice3D<T,Descriptor>& lattice, ScalarField3D<int>& flag)
{
    using namespace descriptors;
    enum { nodeNature = Descriptor<T>::ExternalField::nodeNatureBeginsAt,
           time = Descriptor<T>::ExternalField::timeBeginsAt
    };

    //Dot3D offLF  = computeRelativeDisplacement(lattice,flag);    // lattice-flag
    //Dot3D absoluteOffset = lattice.getLocation();

    for (plint iX=domain.x0; iX<=domain.x1; ++iX) {
        for (plint iY=domain.y0; iY<=domain.y1; ++iY) {
            for (plint iZ=domain.z0; iZ<=domain.z1; ++iZ) {
                if (lattice.get(iX,iY,iZ).getDynamics().getId() !=
                    meta::dynamicsRegistration<T,Descriptor>().getId("BounceBack"))
		        {
		            bool isNextToWall = false;
                    for (plint iPop = 1; iPop < descriptors::D3Q27Descriptor<T>::q; ++iPop)// !! D3Q27 --> Descriptor
                    {
                        plint nextX = iX + descriptors::D3Q27Descriptor<T>::c[iPop][0];
                        plint nextY = iY + descriptors::D3Q27Descriptor<T>::c[iPop][1];
                        plint nextZ = iZ + descriptors::D3Q27Descriptor<T>::c[iPop][2];

                        if (lattice.get(nextX,nextY,nextZ).getDynamics().getId() ==
                            meta::dynamicsRegistration<T,Descriptor>().getId("BounceBack"))
                        {

                            isNextToWall = true;
                            break;
                        }
                    }

                    if (isNextToWall)
                    {
                        *(lattice.get(iX,iY,iZ).getExternal(nodeNature)) = NodeNature::nextToWall;
                    }
                    else *(lattice.get(iX,iY,iZ).getExternal(nodeNature)) = NodeNature::bulk;
                }
            }
        }
    }
}

template<typename T, template<class U> class Descriptor>
DefineBulkStatusFunctional3D<T,Descriptor>* DefineBulkStatusFunctional3D<T,Descriptor>::clone() const {
    return new DefineBulkStatusFunctional3D<T,Descriptor>(*this);
}

template<typename T, template<class U> class Descriptor>
void DefineBulkStatusFunctional3D <T,Descriptor>::getTypeOfModification(std::vector<modif::ModifT>& modified) const {
    modified[0] = modif::staticVariables;
    modified[1] = modif::nothing;
}

////////////////////////




//////////////////////////////////////

template<typename T, template<class U> class Descriptor>
IniWallFlagsFunctional3D<T,Descriptor>::
        IniWallFlagsFunctional3D(plint erodableWallLimit_):
        erodableWallLimit(erodableWallLimit_)
{ }


template<typename T, template<class U> class Descriptor>
void IniWallFlagsFunctional3D<T,Descriptor>::process(
    Box3D domain, BlockLattice3D<T,Descriptor>& lattice, ScalarField3D<int>& flag)
{
    using namespace descriptors;
    enum { nodeNature = Descriptor<T>::ExternalField::nodeNatureBeginsAt,
           time = Descriptor<T>::ExternalField::timeBeginsAt
    };

    //Dot3D offLF  = computeRelativeDisplacement(lattice,flag);    // lattice-flag
    // Dot3D absoluteOffset = lattice.getLocation();

    for (plint iX=domain.x0; iX<=domain.x1; ++iX) {
        for (plint iY=domain.y0; iY<=domain.y1; ++iY) {
            for (plint iZ=domain.z0; iZ<=domain.z1; ++iZ) {

                //plint absoluteZ = absoluteOffset.z + iZ;

                if (lattice.get(iX,iY,iZ).getDynamics().getId() ==
                    meta::dynamicsRegistration<T,Descriptor>().getId("BounceBack"))
                {
//                     if (absoluteZ <= erodableWallLimit)
//                         *(lattice.get(iX,iY,iZ).getExternal(nodeNature)) = NodeNature::erodableWall;
//                     else
                        *(lattice.get(iX,iY,iZ).getExternal(nodeNature)) = NodeNature::notErodableWall;
                }
                else {
                    bool isNextToWall = false;
                    for (plint iPop = 1; iPop < descriptors::D3Q27Descriptor<T>::q; ++iPop) {
                        plint nextX = iX + descriptors::D3Q27Descriptor<T>::c[iPop][0];
                        plint nextY = iY + descriptors::D3Q27Descriptor<T>::c[iPop][1];
                        plint nextZ = iZ + descriptors::D3Q27Descriptor<T>::c[iPop][2];

                        if (lattice.get(nextX,nextY,nextZ).getDynamics().getId() ==
                            meta::dynamicsRegistration<T,Descriptor>().getId("BounceBack")) {

                            isNextToWall = true;
                            break;
                        }
                    }

                    if (isNextToWall) {
                        *(lattice.get(iX,iY,iZ).getExternal(nodeNature)) = NodeNature::nextToWall;
                    }
                    else *(lattice.get(iX,iY,iZ).getExternal(nodeNature)) = NodeNature::bulk;
                }

            }
        }
    }
}

template<typename T, template<class U> class Descriptor>
IniWallFlagsFunctional3D<T,Descriptor>* IniWallFlagsFunctional3D<T,Descriptor>::clone() const {
    return new IniWallFlagsFunctional3D<T,Descriptor>(*this);
}

template<typename T, template<class U> class Descriptor>
void IniWallFlagsFunctional3D<T,Descriptor>::getTypeOfModification(std::vector<modif::ModifT>& modified) const {
    modified[0] = modif::staticVariables;
    modified[1] = modif::nothing;
}

//////////////////////////



template<typename T, template<class U> class Descriptor>
IniWallFlagsCylinderTestFunctional3D<T,Descriptor>::
        IniWallFlagsCylinderTestFunctional3D(plint centerX_, plint centerZ_, plint radius_):
        centerX(centerX_), centerZ(centerZ_), radius(radius_)
{ }


template<typename T, template<class U> class Descriptor>
void IniWallFlagsCylinderTestFunctional3D<T,Descriptor>::process(
    Box3D domain, BlockLattice3D<T,Descriptor>& lattice, ScalarField3D<int>& flag)
{
    using namespace descriptors;
    enum { nodeNature   = Descriptor<T>::ExternalField::nodeNatureBeginsAt,
           time         = Descriptor<T>::ExternalField::timeBeginsAt
    };

    // Dot3D offLF  = computeRelativeDisplacement(lattice,flag);    // lattice-flag

    Dot3D absoluteOffset = lattice.getLocation();

    for (plint iX=domain.x0; iX<=domain.x1; ++iX) {
        plint iX_abs = iX + absoluteOffset.x;

        for (plint iY=domain.y0; iY<=domain.y1; ++iY) {

            for (plint iZ=domain.z0; iZ<=domain.z1; ++iZ) {
                plint iZ_abs = iZ + absoluteOffset.z;



                bool isErodableWall = true;
                if(util::sqr(iX_abs-centerX) + util::sqr(iZ_abs-centerZ) <= radius*radius)
                {
                    isErodableWall = false;
                }



                if (lattice.get(iX,iY,iZ).getDynamics().getId() ==
                    meta::dynamicsRegistration<T,Descriptor>().getId("BounceBack")) {
                    if (isErodableWall)
                        *(lattice.get(iX,iY,iZ).getExternal(nodeNature)) = NodeNature::erodableWall;
                    else
                        *(lattice.get(iX,iY,iZ).getExternal(nodeNature)) = NodeNature::notErodableWall;
                }
                else {
                    bool isNextToWall = false;
                    for (plint iPop = 1; iPop < descriptors::D3Q27Descriptor<T>::q; ++iPop) {
                        plint nextX = iX + descriptors::D3Q27Descriptor<T>::c[iPop][0];
                        plint nextY = iY + descriptors::D3Q27Descriptor<T>::c[iPop][1];
                        plint nextZ = iZ + descriptors::D3Q27Descriptor<T>::c[iPop][2];

                        if (lattice.get(nextX,nextY,nextZ).getDynamics().getId() ==
                            meta::dynamicsRegistration<T,Descriptor>().getId("BounceBack")) {

                            isNextToWall = true;
                            break;
                        }
                    }

                    if (isNextToWall) {
                        *(lattice.get(iX,iY,iZ).getExternal(nodeNature)) = NodeNature::nextToWall;
                    }
                    else *(lattice.get(iX,iY,iZ).getExternal(nodeNature)) = NodeNature::bulk;
                }

            }
        }
    }
}

template<typename T, template<class U> class Descriptor>
IniWallFlagsCylinderTestFunctional3D<T,Descriptor>* IniWallFlagsCylinderTestFunctional3D<T,Descriptor>::clone() const {
    return new IniWallFlagsCylinderTestFunctional3D<T,Descriptor>(*this);
}

template<typename T, template<class U> class Descriptor>
void IniWallFlagsCylinderTestFunctional3D<T,Descriptor>::getTypeOfModification(std::vector<modif::ModifT>& modified) const {
    modified[0] = modif::staticVariables;
    modified[1] = modif::nothing;
}








template<typename T, template<class U> class Descriptor>
IniWallFlagsDamBreakTestFunctional3D<T,Descriptor>::
        IniWallFlagsDamBreakTestFunctional3D(plint centerX_, plint centerZ_, plint radius_):
        centerX(centerX_), centerZ(centerZ_), radius(radius_)
{ }


template<typename T, template<class U> class Descriptor>
void IniWallFlagsDamBreakTestFunctional3D<T,Descriptor>::process(
    Box3D domain, BlockLattice3D<T,Descriptor>& lattice, ScalarField3D<int>& flag)
{
    using namespace descriptors;
    enum { nodeNature   = Descriptor<T>::ExternalField::nodeNatureBeginsAt,
           time         = Descriptor<T>::ExternalField::timeBeginsAt
    };

    //Dot3D offLF  = computeRelativeDisplacement(lattice,flag);    // lattice-flag

    // Dot3D absoluteOffset = lattice.getLocation();

    for (plint iX=domain.x0; iX<=domain.x1; ++iX) {
        //plint iX_abs = iX + absoluteOffset.x;

        for (plint iY=domain.y0; iY<=domain.y1; ++iY) {

            for (plint iZ=domain.z0; iZ<=domain.z1; ++iZ) {
                //plint iZ_abs = iZ + absoluteOffset.z;



                 bool isErodableWall = true;
//                 if(util::sqr(iX_abs-centerX) + util::sqr(iZ_abs-centerZ) <= radius*radius)
//                 {
//                     isErodableWall = false;
//                 }



                if (lattice.get(iX,iY,iZ).getDynamics().getId() ==
                    meta::dynamicsRegistration<T,Descriptor>().getId("BounceBack")) {
                    if (isErodableWall)
                        *(lattice.get(iX,iY,iZ).getExternal(nodeNature)) = NodeNature::erodableWall;
                    else
                        *(lattice.get(iX,iY,iZ).getExternal(nodeNature)) = NodeNature::notErodableWall;
                }
                else {
                    bool isNextToWall = false;
                    for (plint iPop = 1; iPop < descriptors::D3Q27Descriptor<T>::q; ++iPop) {
                        plint nextX = iX + descriptors::D3Q27Descriptor<T>::c[iPop][0];
                        plint nextY = iY + descriptors::D3Q27Descriptor<T>::c[iPop][1];
                        plint nextZ = iZ + descriptors::D3Q27Descriptor<T>::c[iPop][2];

                        if (lattice.get(nextX,nextY,nextZ).getDynamics().getId() ==
                            meta::dynamicsRegistration<T,Descriptor>().getId("BounceBack")) {

                            isNextToWall = true;
                            break;
                        }
                    }

                    if (isNextToWall) {
                        *(lattice.get(iX,iY,iZ).getExternal(nodeNature)) = NodeNature::nextToWall;
                    }
                    else *(lattice.get(iX,iY,iZ).getExternal(nodeNature)) = NodeNature::bulk;
                }

            }
        }
    }
}

template<typename T, template<class U> class Descriptor>
IniWallFlagsDamBreakTestFunctional3D<T,Descriptor>* IniWallFlagsDamBreakTestFunctional3D<T,Descriptor>::clone() const {
    return new IniWallFlagsDamBreakTestFunctional3D<T,Descriptor>(*this);
}

template<typename T, template<class U> class Descriptor>
void IniWallFlagsDamBreakTestFunctional3D<T,Descriptor>::getTypeOfModification(std::vector<modif::ModifT>& modified) const {
    modified[0] = modif::staticVariables;
    modified[1] = modif::nothing;
}





template<typename T, template<class U> class Descriptor>
void IniWallFlagRiverTestFunctional3D<T,Descriptor>::process(
    Box3D domain, BlockLattice3D<T,Descriptor>& lattice, ScalarField3D<int>& flag)
{
    using namespace descriptors;
    enum { nodeNature   = Descriptor<T>::ExternalField::nodeNatureBeginsAt,
           time         = Descriptor<T>::ExternalField::timeBeginsAt};

    //Dot3D offLF  = computeRelativeDisplacement(lattice,flag);    // lattice-flag

    //Dot3D absoluteOffset = lattice.getLocation();



    for (plint iX=domain.x0; iX<=domain.x1; ++iX) {
//         plint iX_abs = iX + absoluteOffset.x;

        for (plint iY=domain.y0; iY<=domain.y1; ++iY) {

            for (plint iZ=domain.z0; iZ<=domain.z1; ++iZ) {
//                 plint iZ_abs = iZ + absoluteOffset.z;




//                 if (lattice.get(iX,iY,iZ).getDynamics().getId() ==
//                     meta::dynamicsRegistration<T,Descriptor>().getId("BounceBack")) {
//                     if ( *(lattice.get(iX,iY,iZ).getExternal(nodeNature)) != NodeNature::notErodableWall )
//                          *(lattice.get(iX,iY,iZ).getExternal(nodeNature)) = NodeNature::erodableWall;
// //                      else
// //                          *(lattice.get(iX,iY,iZ).getExternal(nodeNature)) = NodeNature::notErodableWall;
//                 }
                if (lattice.get(iX,iY,iZ).getDynamics().getId() !=
                    meta::dynamicsRegistration<T,Descriptor>().getId("BounceBack"))
                {

                    bool isNextToWall = false;
                    for (plint iPop = 1; iPop < descriptors::D3Q27Descriptor<T>::q; ++iPop)
                    {
                        plint nextX = iX + descriptors::D3Q27Descriptor<T>::c[iPop][0];
                        plint nextY = iY + descriptors::D3Q27Descriptor<T>::c[iPop][1];
                        plint nextZ = iZ + descriptors::D3Q27Descriptor<T>::c[iPop][2];

                        if (lattice.get(nextX,nextY,nextZ).getDynamics().getId() ==
                            meta::dynamicsRegistration<T,Descriptor>().getId("BounceBack"))
                            {
                                isNextToWall = true;
                                break;
                            }
                    }

                    if (isNextToWall)
                    {
                        *(lattice.get(iX,iY,iZ).getExternal(nodeNature)) = NodeNature::nextToWall;
                    }
                    else
                    {
			            *(lattice.get(iX,iY,iZ).getExternal(nodeNature)) = NodeNature::bulk;
		            }
		        }

            }
        }
    }
}

template<typename T, template<class U> class Descriptor>
IniWallFlagRiverTestFunctional3D<T,Descriptor>* IniWallFlagRiverTestFunctional3D<T,Descriptor>::clone() const {
    return new IniWallFlagRiverTestFunctional3D<T,Descriptor>(*this);
}

template<typename T, template<class U> class Descriptor>
void IniWallFlagRiverTestFunctional3D<T,Descriptor>::getTypeOfModification(std::vector<modif::ModifT>& modified) const {
    modified[0] = modif::staticVariables;
    modified[1] = modif::nothing;
}






















/////////////////////////////////////////////



template<typename T1, template<class U> class Descriptor1, typename T2, template<class U> class Descriptor2>
IniDynamicsForParticlesByUsingLatticeFunctional3D<T1,Descriptor1,T2,Descriptor2>::
        IniDynamicsForParticlesByUsingLatticeFunctional3D(Dynamics<T1,Descriptor1>* dynamics_):
	dynamics(dynamics_)
{ }



template<typename T1, template<class U> class Descriptor1, typename T2, template<class U> class Descriptor2>
void IniDynamicsForParticlesByUsingLatticeFunctional3D<T1,Descriptor1,T2,Descriptor2>::process(
    Box3D domain, BlockLattice3D<T1,Descriptor1>& particles, BlockLattice3D<T2,Descriptor2>& lattice)
{
    using namespace descriptors;
    enum { nodeNature 	= Descriptor2<T2>::ExternalField::nodeNatureBeginsAt };

    Dot3D offset  = computeRelativeDisplacement(particles,lattice);    // lattice-flag

    for (plint iX=domain.x0; iX<=domain.x1; ++iX) {
	plint oX = iX + offset.x ;
        for (plint iY=domain.y0; iY<=domain.y1; ++iY) {
            plint oY = iY + offset.y ;
	    for (plint iZ=domain.z0; iZ<=domain.z1; ++iZ) {
                plint oZ = iZ + offset.z ;

		// when I enter de dataProcessor all the particles nodes are AdvectionDiffusionDynamics

		if ( lattice.get(oX,oY,oZ).getDynamics().getId() ==
                            meta::dynamicsRegistration<T2,Descriptor2>().getId("BounceBack") )
		{
		    bool bulkWallNode = true;

		    for (plint iPop = 1; iPop < descriptors::D3Q27Descriptor<T1>::q; ++iPop)
		    {

			plint nextX_l = oX + descriptors::D3Q27Descriptor<T1>::c[iPop][0];
			plint nextY_l = oY + descriptors::D3Q27Descriptor<T1>::c[iPop][1];
			plint nextZ_l = oZ + descriptors::D3Q27Descriptor<T1>::c[iPop][2];


			if ( *(lattice.get(nextX_l,nextY_l,nextZ_l).getExternal(nodeNature)) == NodeNature::nextToWall )
			{
			    bulkWallNode = false;
			    //particles.attributeDynamics(iX,iY,iZ, dynamics->clone() );
			    break;
			}

		    }
		    if (bulkWallNode)
			particles.attributeDynamics(iX,iY,iZ, new NoDynamics<T1,Descriptor1>() );

		}

// 		if (lattice.get(oX,oY,oZ).getDynamics().getId() ==
//                             meta::dynamicsRegistration<T2,Descriptor2>().getId("NoDynamics")  )
//                 {
//                     particles.attributeDynamics(iX,iY,iZ, new NoDynamics<T1,Descriptor1>());
//                 }


	    }
        }
    }
}

template<typename T1, template<class U> class Descriptor1, typename T2, template<class U> class Descriptor2>
IniDynamicsForParticlesByUsingLatticeFunctional3D<T1,Descriptor1,T2,Descriptor2>* IniDynamicsForParticlesByUsingLatticeFunctional3D<T1,Descriptor1,T2,Descriptor2>::clone() const {
    return new IniDynamicsForParticlesByUsingLatticeFunctional3D<T1,Descriptor1,T2,Descriptor2>(*this);
}

template<typename T1, template<class U> class Descriptor1, typename T2, template<class U> class Descriptor2>
void IniDynamicsForParticlesByUsingLatticeFunctional3D<T1,Descriptor1,T2,Descriptor2>::getTypeOfModification(std::vector<modif::ModifT>& modified) const {
    modified[0] = modif::staticVariables;
    modified[1] = modif::nothing;
}






template<typename T, template<class U> class Descriptor>
IniDynamicsPerRiverFunctional3D<T,Descriptor>::
        IniDynamicsPerRiverFunctional3D(Dynamics<T,Descriptor>* dynamics_):
        dynamics(dynamics_)
{ }



template<typename T, template<class U> class Descriptor>
void IniDynamicsPerRiverFunctional3D<T,Descriptor>::process(Box3D domain,
                                                            BlockLattice3D<T,Descriptor>& particles,
                                                            ScalarField3D<int>& flagFluid)
{
    using namespace descriptors;
    using namespace freeSurfaceFlag;

    Dot3D offset  = computeRelativeDisplacement(particles,flagFluid);    // lattice-flag

    for (plint iX=domain.x0; iX<=domain.x1; ++iX) {
        plint oX = iX + offset.x ;
        for (plint iY=domain.y0; iY<=domain.y1; ++iY) {
            plint oY = iY + offset.y ;
            for (plint iZ=domain.z0; iZ<=domain.z1; ++iZ) {
                plint oZ = iZ + offset.z ;

                // when entering the dataProcessor all the particles nodes are AdvectionDiffusionDynamics

                if (flagFluid.get(oX,oY,oZ) == freeSurfaceFlag::fluid ||
                    flagFluid.get(oX,oY,oZ) == freeSurfaceFlag::interface)
                {
                    particles.attributeDynamics(iX,iY,iZ, dynamics->clone() );
                }
                else if (flagFluid.get(oX,oY,oZ) == freeSurfaceFlag::wall)
                {
                    particles.attributeDynamics(iX,iY,iZ, new NoDynamics<T,Descriptor>() );
                }

            }
        }
    }
}

template<typename T, template<class U> class Descriptor>
IniDynamicsPerRiverFunctional3D<T,Descriptor>* IniDynamicsPerRiverFunctional3D<T,Descriptor>::clone() const {
    return new IniDynamicsPerRiverFunctional3D<T,Descriptor>(*this);
}

template<typename T, template<class U> class Descriptor>
void IniDynamicsPerRiverFunctional3D<T,Descriptor>::getTypeOfModification(std::vector<modif::ModifT>& modified) const {
    modified[0] = modif::dataStructure;
    modified[1] = modif::nothing;
}



}  // namespace plb



#endif  // DATA_ANALYSIS_FUNCTIONAL_3D_HH
