/* This file is part of the Palabos library.
 * Copyright (C) 2009 Jonas Latt
 * E-mail contact: jonas@lbmethod.org
 * The most recent release of Palabos can be downloaded at 
 * <http://www.lbmethod.org/palabos/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/** \file
 * Descriptor for nearest-neighbor 3D lattices. In principle, thanks
 * to static genericity of the code, definition of the lattice and
 * implementation of the dynamics are independent. Of course, there
 * remain a few dependencies. For example, a dynamics which uses higher-
 * order moments such as heat flux needs a lattice with an extended
 * neighborhood. Another example is the D3Q13 lattice, which works only
 * with its very own dynamics.
 *
 *  -- header file
 */
#ifndef THROMBOSIS_LATTICES_3D_HJJ
#define THROMBOSIS_LATTICES_3D_HJJ

#include "core/globalDefs.h"
#include "core/plbDebug.h"
#include "latticeBoltzmann/nearestNeighborLattices3D.h"
#include "latticeBoltzmann/nearestNeighborLattices3D.hh"
#include "latticeBoltzmann/externalFields.h"
#include "latticeBoltzmann/roundOffPolicy.h"
#include <vector>

namespace plb {

/// Descriptors for the 2D and 3D lattices.
/** \warning Attention: The lattice directions must always be ordered in
 * such a way that c[i] = -c[i+(q-1)/2] for i=1..(q-1)/2, and c[0] = 0 must
 * be the rest velocity. Furthermore, the velocities c[i] for i=1..(q-1)/2
 * must verify
 *  - in 2D: (c[i][0]<0) || (c[i][0]==0 && c[i][1]<0)
 *  - in 3D: (c[i][0]<0) || (c[i][0]==0 && c[i][1]<0)
 *                       || (c[i][0]==0 && c[i][1]==0 && c[i][2]<0)
 * Otherwise some of the code will work erroneously, because the
 * aformentioned relations are taken as given to enable a few
 * optimizations.
*/
namespace descriptors {

   
    //////////////////
    
     struct FreeSurfaceSedimentationDescriptor {
        static const int numScalars = 7;
        static const int numSpecies = 5;
        
        static const int nodeNatureBeginsAt = 0;
        static const int sizeOfNodeNature   = 1;
        static const int timeBeginsAt = 1; // for deposition Process
        static const int sizeOfTime   = 1;
        static const int timeErosionBeginsAt = 2; // for Erosion Process
        static const int sizeOfTimeErosion   = 1;
	static const int omegaBeginsAt = 3;
	static const int sizeOfOmega = 1;
	
        static const int forceBeginsAt    = 4;
        static const int sizeOfForce      = 3;
	
	
	
    };
    
    struct FreeSurfaceSedimentationDescriptorBase {
        typedef FreeSurfaceSedimentationDescriptor ExternalField;
    };

    template <typename T> struct FreeSurfaceSedimentationD3Q19Descriptor
        : public D3Q19DescriptorBase<T>, public FreeSurfaceSedimentationDescriptorBase
    { 
        static const char name[];
    };
    
    
    
    
}  // namespace descriptors

}  // namespace plb

#endif
